@@ -1,16 +1,15 @@
 """
 Python Environment API for the ML-Agents toolkit
-The aim of this API is to expose groups of similar Agents evolving in Unity
+The aim of this API is to expose Agents evolving in a simulation
 to perform reinforcement learning on.
-There can be multiple groups of similar Agents (same observations and actions
-spaces) in the simulation. These groups are identified by a agent_group that
-corresponds to a single group of Agents in the simulation.
+This API supports multi-agent scenarios and groups similar Agents (same
+observations, actions spaces and behavior) together. These groups of Agents are
+identified by their BehaviorName.
 For performance reasons, the data of each group of agents is processed in a
-batched manner. When retrieving the state of a group of Agents, said state
-contains the data for the whole group. Agents in these groups are identified
-by a unique int identifier that allows tracking of Agents across simulation
-steps. Note that there is no guarantee that the number or order of the Agents
-in the state will be consistent across simulation steps.
+batched manner. Agents are identified by a unique AgentId identifier that
+allows tracking of Agents across simulation steps. Note that there is no
+guarantee that the number or order of the Agents in the state will be
+consistent across simulation steps.
 A simulation steps corresponds to moving the simulation forward until at least
 one agent in the simulation sends its observations to Python again. Since
 Agents can request decisions at different frequencies, a simulation step does
@@ -18,26 +17,22 @@ not necessarily correspond to a fixed simulation time increment.
 """
 
 from abc import ABC, abstractmethod
-from typing import List, NamedTuple, Tuple, Optional, Union, Dict
+from collections.abc import Mapping
+from typing import List, NamedTuple, Tuple, Optional, Union, Dict, Iterator, Any
 import numpy as np
 from enum import Enum
 
 AgentId = int
-AgentGroup = str
+BehaviorName = str
 
 
-class StepResult(NamedTuple):
+class DecisionStep(NamedTuple):
     """
     Contains the data a single Agent collected since the last
     simulation step.
-     - obs is a list of numpy arrays observations collected by the group of
-     agent.
+     - obs is a list of numpy arrays observations collected by the agent.
      - reward is a float. Corresponds to the rewards collected by the agent
      since the last simulation step.
-     - done is a bool. Is true if the Agent was terminated during the last
-     simulation step.
-     - max_step is a bool. Is true if the Agent reached its maximum number of
-     steps during the last simulation step.
      - agent_id is an int and an unique identifier for the corresponding Agent.
      - action_mask is an optional list of one dimensional array of booleans.
      Only available in multi-discrete action space type.
@@ -48,30 +43,22 @@ class StepResult(NamedTuple):
 
     obs: List[np.ndarray]
     reward: float
-    done: bool
-    max_step: bool
     agent_id: AgentId
     action_mask: Optional[List[np.ndarray]]
 
 
-class BatchedStepResult:
+class DecisionSteps(Mapping):
     """
-    Contains the data a group of similar Agents collected since the last
+    Contains the data a batch of similar Agents collected since the last
     simulation step. Note that all Agents do not necessarily have new
     information to send at each simulation step. Therefore, the ordering of
-    agents and the batch size of the BatchedStepResult are not fixed across
+    agents and the batch size of the DecisionSteps are not fixed across
     simulation steps.
-     - obs is a list of numpy arrays observations collected by the group of
-     agent. Each obs has one extra dimension compared to StepResult: the first
-     dimension of the array corresponds to the batch size of
-     the group.
+     - obs is a list of numpy arrays observations collected by the batch of
+     agent. Each obs has one extra dimension compared to DecisionStep: the
+     first dimension of the array corresponds to the batch size of the batch.
      - reward is a float vector of length batch size. Corresponds to the
      rewards collected by each agent since the last simulation step.
-     - done is an array of booleans of length batch size. Is true if the
-     associated Agent was terminated during the last simulation step.
-     - max_step is an array of booleans of length batch size. Is true if the
-     associated Agent reached its maximum number of steps during the last
-     simulation step.
      - agent_id is an int vector of length batch size containing unique
      identifier for the corresponding Agent. This is used to track Agents
      across simulation steps.
@@ -83,11 +70,9 @@ class BatchedStepResult:
      this simulation step.
     """
 
-    def __init__(self, obs, reward, done, max_step, agent_id, action_mask):
+    def __init__(self, obs, reward, agent_id, action_mask):
         self.obs: List[np.ndarray] = obs
         self.reward: np.ndarray = reward
-        self.done: np.ndarray = done
-        self.max_step: np.ndarray = max_step
         self.agent_id: np.ndarray = agent_id
         self.action_mask: Optional[List[np.ndarray]] = action_mask
         self._agent_id_to_index: Optional[Dict[AgentId, int]] = None
@@ -95,10 +80,8 @@ class BatchedStepResult:
     @property
     def agent_id_to_index(self) -> Dict[AgentId, int]:
         """
-        Returns the index of the agent_id in this BatchedStepResult, and
-        -1 if agent_id is not in this BatchedStepResult.
-        :param agent_id: The id of the agent
-        :returns: The index of the agent_id, and -1 if not found.
+        :returns: A Dict that maps agent_id to the index of those agents in
+        this DecisionSteps.
         """
         if self._agent_id_to_index is None:
             self._agent_id_to_index = {}
@@ -106,19 +89,18 @@ class BatchedStepResult:
                 self._agent_id_to_index[a_id] = a_idx
         return self._agent_id_to_index
 
-    def contains_agent(self, agent_id: AgentId) -> bool:
-        return agent_id in self.agent_id_to_index
+    def __len__(self) -> int:
+        return len(self.agent_id)
 
-    def get_agent_step_result(self, agent_id: AgentId) -> StepResult:
+    def __getitem__(self, agent_id: AgentId) -> DecisionStep:
         """
-        returns the step result for a specific agent.
+        returns the DecisionStep for a specific agent.
         :param agent_id: The id of the agent
-        :returns: obs, reward, done, agent_id and optional action mask for a
-        specific agent
+        :returns: The DecisionStep
         """
-        if not self.contains_agent(agent_id):
-            raise IndexError(
-                "agent_id {} is not present in the BatchedStepResult".format(agent_id)
+        if agent_id not in self.agent_id_to_index:
+            raise KeyError(
+                "agent_id {} is not present in the DecisionSteps".format(agent_id)
             )
         agent_index = self._agent_id_to_index[agent_id]  # type: ignore
         agent_obs = []
@@ -129,49 +111,144 @@ class BatchedStepResult:
             agent_mask = []
             for mask in self.action_mask:
                 agent_mask.append(mask[agent_index])
-        return StepResult(
+        return DecisionStep(
             obs=agent_obs,
             reward=self.reward[agent_index],
-            done=self.done[agent_index],
-            max_step=self.max_step[agent_index],
             agent_id=agent_id,
             action_mask=agent_mask,
         )
 
+    def __iter__(self) -> Iterator[Any]:
+        yield from self.agent_id
+
     @staticmethod
-    def empty(spec: "AgentGroupSpec") -> "BatchedStepResult":
+    def empty(spec: "BehaviorSpec") -> "DecisionSteps":
         """
-        Returns an empty BatchedStepResult.
-        :param spec: The AgentGroupSpec for the BatchedStepResult
+        Returns an empty DecisionSteps.
+        :param spec: The BehaviorSpec for the DecisionSteps
         """
         obs: List[np.ndarray] = []
         for shape in spec.observation_shapes:
             obs += [np.zeros((0,) + shape, dtype=np.float32)]
-        return BatchedStepResult(
+        return DecisionSteps(
             obs=obs,
             reward=np.zeros(0, dtype=np.float32),
-            done=np.zeros(0, dtype=np.bool),
-            max_step=np.zeros(0, dtype=np.bool),
             agent_id=np.zeros(0, dtype=np.int32),
             action_mask=None,
         )
 
-    def n_agents(self) -> int:
+
+class TerminalStep(NamedTuple):
+    """
+    Contains the data a single Agent collected when its episode ended.
+     - obs is a list of numpy arrays observations collected by the agent.
+     - reward is a float. Corresponds to the rewards collected by the agent
+     since the last simulation step.
+     - max_step is a bool. Is true if the Agent reached its maximum number of
+     steps during the last simulation step.
+     - agent_id is an int and an unique identifier for the corresponding Agent.
+    """
+
+    obs: List[np.ndarray]
+    reward: float
+    max_step: bool
+    agent_id: AgentId
+
+
+class TerminalSteps(Mapping):
+    """
+    Contains the data a batch of Agents collected when their episode
+    terminated. All Agents present in the TerminalSteps have ended their
+    episode.
+     - obs is a list of numpy arrays observations collected by the batch of
+     agent. Each obs has one extra dimension compared to DecisionStep: the
+     first dimension of the array corresponds to the batch size of the batch.
+     - reward is a float vector of length batch size. Corresponds to the
+     rewards collected by each agent since the last simulation step.
+     - max_step is an array of booleans of length batch size. Is true if the
+     associated Agent reached its maximum number of steps during the last
+     simulation step.
+     - agent_id is an int vector of length batch size containing unique
+     identifier for the corresponding Agent. This is used to track Agents
+     across simulation steps.
+    """
+
+    def __init__(self, obs, reward, max_step, agent_id):
+        self.obs: List[np.ndarray] = obs
+        self.reward: np.ndarray = reward
+        self.max_step: np.ndarray = max_step
+        self.agent_id: np.ndarray = agent_id
+        self._agent_id_to_index: Optional[Dict[AgentId, int]] = None
+
+    @property
+    def agent_id_to_index(self) -> Dict[AgentId, int]:
+        """
+        :returns: A Dict that maps agent_id to the index of those agents in
+        this TerminalSteps.
+        """
+        if self._agent_id_to_index is None:
+            self._agent_id_to_index = {}
+            for a_idx, a_id in enumerate(self.agent_id):
+                self._agent_id_to_index[a_id] = a_idx
+        return self._agent_id_to_index
+
+    def __len__(self) -> int:
         return len(self.agent_id)
 
+    def __getitem__(self, agent_id: AgentId) -> TerminalStep:
+        """
+        returns the TerminalStep for a specific agent.
+        :param agent_id: The id of the agent
+        :returns: obs, reward, done, agent_id and optional action mask for a
+        specific agent
+        """
+        if agent_id not in self.agent_id_to_index:
+            raise KeyError(
+                "agent_id {} is not present in the TerminalSteps".format(agent_id)
+            )
+        agent_index = self._agent_id_to_index[agent_id]  # type: ignore
+        agent_obs = []
+        for batched_obs in self.obs:
+            agent_obs.append(batched_obs[agent_index])
+        return TerminalStep(
+            obs=agent_obs,
+            reward=self.reward[agent_index],
+            max_step=self.max_step[agent_index],
+            agent_id=agent_id,
+        )
+
+    def __iter__(self) -> Iterator[Any]:
+        yield from self.agent_id
+
+    @staticmethod
+    def empty(spec: "BehaviorSpec") -> "TerminalSteps":
+        """
+        Returns an empty TerminalSteps.
+        :param spec: The BehaviorSpec for the TerminalSteps
+        """
+        obs: List[np.ndarray] = []
+        for shape in spec.observation_shapes:
+            obs += [np.zeros((0,) + shape, dtype=np.float32)]
+        return TerminalSteps(
+            obs=obs,
+            reward=np.zeros(0, dtype=np.float32),
+            max_step=np.zeros(0, dtype=np.bool),
+            agent_id=np.zeros(0, dtype=np.int32),
+        )
+
 
 class ActionType(Enum):
     DISCRETE = 0
     CONTINUOUS = 1
 
 
-class AgentGroupSpec(NamedTuple):
+class BehaviorSpec(NamedTuple):
     """
     A NamedTuple to containing information about the observations and actions
-    spaces for a group of Agents.
+    spaces for a group of Agents under the same behavior.
      - observation_shapes is a List of Tuples of int : Each Tuple corresponds
      to an observation's dimensions. The shape tuples have the same ordering as
-     the ordering of the BatchedStepResult and StepResult.
+     the ordering of the DecisionSteps and TerminalSteps.
      - action_type is the type of data of the action. it can be discrete or
      continuous. If discrete, the action tensors are expected to be int32. If
      continuous, the actions are expected to be float32.
@@ -188,13 +265,13 @@ class AgentGroupSpec(NamedTuple):
 
     def is_action_discrete(self) -> bool:
         """
-        Returns true if the Agent group uses discrete actions
+        Returns true if this Behavior uses discrete actions
         """
         return self.action_type == ActionType.DISCRETE
 
     def is_action_continuous(self) -> bool:
         """
-        Returns true if the Agent group uses continuous actions
+        Returns true if this Behavior uses continuous actions
         """
         return self.action_type == ActionType.CONTINUOUS
 
@@ -254,23 +331,24 @@ class BaseEnv(ABC):
         pass
 
     @abstractmethod
-    def get_agent_groups(self) -> List[AgentGroup]:
+    def get_behavior_names(self) -> List[BehaviorName]:
         """
-        Returns the list of the agent group names present in the environment.
-        Agents grouped under the same group name have the same action and
-        observation specs, and are expected to behave similarly in the environment.
+        Returns the list of the behavior names present in the environment.
+        Agents grouped under the same behavior name have the same action and
+        observation specs, and are expected to behave similarly in the
+        environment.
         This list can grow with time as new policies are instantiated.
-        :return: the list of agent group names.
+        :return: the list of agent BehaviorName.
         """
         pass
 
     @abstractmethod
-    def set_actions(self, agent_group: AgentGroup, action: np.ndarray) -> None:
+    def set_actions(self, behavior_name: BehaviorName, action: np.ndarray) -> None:
         """
         Sets the action for all of the agents in the simulation for the next
         step. The Actions must be in the same order as the order received in
-        the step result.
-        :param agent_group: The name of the group the agents are part of
+        the DecisionSteps.
+        :param behavior_name: The name of the behavior the agents are part of
         :param action: A two dimensional np.ndarray corresponding to the action
         (either int or float)
         """
@@ -278,34 +356,41 @@ class BaseEnv(ABC):
 
     @abstractmethod
     def set_action_for_agent(
-        self, agent_group: AgentGroup, agent_id: AgentId, action: np.ndarray
+        self, behavior_name: BehaviorName, agent_id: AgentId, action: np.ndarray
     ) -> None:
         """
         Sets the action for one of the agents in the simulation for the next
         step.
-        :param agent_group: The name of the group the agent is part of
+        :param behavior_name: The name of the behavior the agent is part of
         :param agent_id: The id of the agent the action is set for
-        :param action: A two dimensional np.ndarray corresponding to the action
+        :param action: A one dimensional np.ndarray corresponding to the action
         (either int or float)
         """
         pass
 
     @abstractmethod
-    def get_step_result(self, agent_group: AgentGroup) -> BatchedStepResult:
+    def get_steps(
+        self, behavior_name: BehaviorName
+    ) -> Tuple[DecisionSteps, TerminalSteps]:
         """
-        Retrieves the observations of the agents that requested a step in the
+        Retrieves the steps of the agents that requested a step in the
         simulation.
-        :param agent_group: The name of the group the agents are part of
-        :return: A BatchedStepResult NamedTuple containing the observations,
-        the rewards and the done flags for this group of agents.
+        :param behavior_name: The name of the behavior the agents are part of
+        :return: A tuple containing :
+         - A DecisionSteps NamedTuple containing the observations,
+         the rewards, the agent ids and the action masks for the Agents
+         of the specified behavior. These Agents need an action this step.
+         - A TerminalSteps NamedTuple containing the observations,
+         rewards, agent ids and max_step flags of the agents that had their
+         episode terminated last step.
         """
         pass
 
     @abstractmethod
-    def get_agent_group_spec(self, agent_group: AgentGroup) -> AgentGroupSpec:
+    def get_behavior_spec(self, behavior_name: BehaviorName) -> BehaviorSpec:
         """
-        Get the AgentGroupSpec corresponding to the agent group name
-        :param agent_group: The name of the group the agents are part of
-        :return: A AgentGroupSpec corresponding to that agent group name
+        Get the BehaviorSpec corresponding to the behavior name
+        :param behavior_name: The name of the behavior the agents are part of
+        :return: A BehaviorSpec corresponding to that behavior
         """
         pass
